from __future__ import annotations

import keyword
import logging
import os
import re
from dataclasses import dataclass, make_dataclass
from typing import (Annotated, Any, Dict, Iterator, Literal, NewType, Tuple,
                    Union)

import bids.layout
import jsonschema
import openapi_schema_validator.validators
from apischema import discriminator, schema
from apischema.json_schema import deserialization_schema

lgr = logging.getLogger(__name__)
DEBUG = bool(os.environ.get("DEBUG", False))
lgr.setLevel(logging.DEBUG if DEBUG else logging.INFO)

FORBIDS_SCHEMA_FOLDER = ".forbids"

# entities that differentiate files from the same series
# where it might be None for one of the files.
ALT_ENTITIES = ["reconstruction", "acquisition"]


def tagpreset2type(tag: str, tag_preset: str, value: Any):
    # Converts tag presets from config files into a typing with constraints

    # Parameters:
    #   tag: the name of the tag
    #   tag_preset: the expression to set constraint from
    #   value: the exemplar value to determine type and set constants

    # Returns:
    #   type: a python type with added apischema constraints
    if tag_preset == "=":
        if isinstance(value, list):
            return Tuple[*[Literal[vv] for vv in value]]
        else:
            return Literal[value]
    elif tag_preset == "*":
        return type(value)
    elif tag_preset.startswith("~="):
        tag = NewType(tag, type(value))
        tol = float(tag_preset[2:])
        schema(min=value - tol, max=value + tol)(tag)
        return tag
    elif tag_preset.startswith("r"):
        tag = NewType(tag, type(value))  # likely a string, check if this need enforcement
        schema(pattern=re.compile(tag_preset[1:]))(tag)
        return tag
    else:
        raise RuntimeError(f"Unsupported constraint for {tag} in the config file.")


def struct2schemaprops(sidecar: dict, config_props: dict, schema_path: str) -> Iterator:
    # from a example dictionary or list and matching config,
    # generate dataclass fields definitions
    # recursively creates subschema for dictionary values

    # Parameters:
    # sidecar: dict with exemplar values to determine type and constant
    # config_props: dict with preset of how to validate tags
    # schema_path: schema name to be postpended in recursive schema defs for uniqness
    for k, tag_preset in config_props.items():
        if k in sidecar:
            v = sidecar[k]
            k2 = k + ("__" if k in keyword.kwlist else "")
            if isinstance(config_props[k], dict):
                yield k2, sidecar2schema(v, tag_preset, f"{schema_path}_{k}")
            elif isinstance(config_props[k], list):
                yield k2, sidecar2schema(v, tag_preset, f"{schema_path}_{k}")
            else:
                yield k2, tagpreset2type(f"{schema_path}_{k}", tag_preset, v)


def sidecar2schema(sidecar: dict, config_props: dict, subschema_name: str):
    # from a exemplar sidecar and config, generate a schema

    # Parameters:
    # sidecar: exemplar sidecar
    # config_props: schema properties config
    # subschema_name: name to give the subschema
    return make_dataclass(subschema_name, fields=list(struct2schemaprops(sidecar, config_props, subschema_name)))


def get_validator(sidecar_schema: dict) -> jsonschema.validators._Validator:
    # return OpenApi validator for use of discriminator feature
    validator_cls = openapi_schema_validator.validators.OAS31Validator
    validator_cls.check_schema(sidecar_schema)
    # validator_cls = jsonschema.validators.validator_for(sidecar_schema)
    return validator_cls(sidecar_schema)


def sidecars2unionschema(
    sidecars_groups: dict[Any, list[bids.layout.BIDSJSONFile]],
    bids_layout: bids.BIDSLayout,
    config_props: dict,
    series_entities: dict,
    factor_entities: tuple = ("subject", "run"),
) -> Annotated:
    # from a set of grouped sidecars from different scanners generate a meta-schema

    schema_name = bids_layout.build_path(series_entities, absolute_paths=False)[:-5] + "-"

    subschemas = []
    mapping_keys = []
    for keys, sidecars in sidecars_groups.items():
        instrument_tags = [k[0] for k in keys]
        sidecars = list(sidecars)
        # generate schema from first exemplar
        sc = sidecars[0]
        lgr.info("generating schema from %s", sc.relpath)
        metas = prepare_metadata(sc, instrument_tags)
        mapping_keys.append(metas["__instrument__"])
        subschema_name = schema_name + "".join([k.replace(".", "") for t, k in keys])
        subschema_name = subschema_name.replace("_", "").replace("-", "")
        # while subschema_name in subschemas:
        #    subschema_name = subschema_name + "_copy"
        subschema = sidecar2schema(metas, config_props, subschema_name)
        # check if we can apply the schema from 1st sidecar to the others:
        validator = get_validator(deserialization_schema(subschema, additional_properties=True))
        for sidecar in sidecars[1:]:
            lgr.info("validating schema from %s", sidecar.relpath)
            # validate or raise
            validator.validate(prepare_metadata(sidecar, instrument_tags))

        subschemas.append(subschema)

#    # if homogeneous (eg. single-site or single-vendor)
#    if len(subschemas) == 1:
#        return subschemas[0]

    UnionModel = Annotated[
        Union[tuple(subschemas)],
        discriminator(
            "__instrument__",
            #            {k :sc.__name__ for k, sc in zip(mapping_keys, subschemas)}
        ),
    ]
    print(UnionModel)

    return UnionModel


def compare_schema(sc1: dataclass, sc2: dataclass) -> bool:
    # compares 2 dataclasses, not really useful now
    match = True

    sc1_props = sc1.__dataclass_fields__
    sc2_props = sc2.__dataclass_fields__
    sc1_props_keys = set(sc1_props.keys())
    sc2_props_keys = set(sc2_props.keys())
    lgr.debug("XOR: %s", str(set(sc1_props_keys).symmetric_difference(sc2_props_keys)))
    for prop in sc1_props_keys.intersection(sc2_props_keys):
        if isinstance(sc1_props[prop], type):
            match = compare_schema(sc1_props[prop], sc2_props[prop])
        t1, t2 = sc1_props[prop].type, sc2_props[prop].type
        if t1 != t2:
            if not hasattr(t1, "__supertype__") or t1.__supertype__ != t2.__supertype__:
                lgr.debug(str((prop, sc1_props[prop].type, sc2_props[prop].type)))
                match = False
    return match


def get_instrument_key(
    sidecar_data: dict[str, Any],
    instrument_tags: List[str],
):
    # composes a instrument key from the metadata and tags list
    return "-".join([sidecar_data.get(instr_tag, "unknown") for instr_tag in instrument_tags])


def prepare_metadata(
    sidecar: bids.layout.BIDSJSONFile,
    instrument_tags: List[str],
):
    # prepares sidecar data for use with json_schema

    # rename conflictual keywords as the schema was created
    sidecar_data = {k + ("__" if k in keyword.kwlist else ""): v for k, v in sidecar.get_dict().items()}
    # create an aggregate tag of all schema-defined instrument tags
    sidecar_data["__instrument__"] = get_instrument_key(sidecar_data, instrument_tags)
    return sidecar_data



from __future__ import annotations

import argparse
import logging
import os

import bids
import coloredlogs

from ..init import initialize
from ..validation import process_validation

DEBUG = bool(os.environ.get("DEBUG", False))
coloredlogs.install()
if DEBUG:
    logging.basicConfig(level=logging.DEBUG)
    logging.root.setLevel(logging.DEBUG)
    root_handler = logging.root.handlers[0]
    root_handler.setFormatter(
        logging.Formatter("%(asctime)s,%(msecs)03d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s")
    )
else:
    root_handler = logging.root.handlers[0]
    root_handler.setFormatter(logging.Formatter("%(levelname)-8s %(message)s"))
    logging.root.setLevel(logging.INFO)

lgr = logging.getLogger(__name__)


def parse_args() -> argparse.Namespace:

    p = argparse.ArgumentParser(description="forbids - setup and validate protocol compliance")
    p.add_argument("command", help="init or validate")
    p.add_argument("bids_path", help="path to the BIDS dataset")
    p.add_argument(
        "--session-specific",
        action="store_true",
        default=False,
        help="build a different schema for each session, is the study design is not repeated measures",
    )

    p.add_argument(
        "--scanner-specific",
        action="store_true",
        default=False,
        help="allow schema to be scanner instance specific",
    )
    p.add_argument(
        "--version-specific",
        action="store_true",
        default=False,
        help="allow schema to be specific to the scanner software version",
    )
    p.add_argument("--participant-label", nargs="+", default=bids.layout.Query.ANY)
    p.add_argument("--session-label", nargs="*", default=[bids.layout.Query.NONE, bids.layout.Query.ANY])
    return p.parse_args()


def main() -> None:
    args = parse_args()
    layout = bids.BIDSLayout(os.path.abspath(args.bids_path))
    success = False

    if args.command == "init":
        success = initialize(
            layout,
            uniform_sessions=not args.session_specific,
            uniform_instruments=not args.scanner_specific,
            version_specific=args.version_specific,
        )
    elif args.command == "validate":
        success = process_validation(layout, subject=args.participant_label, session=args.session_label)
    exit(0 if success else 1)


if __name__ == "__main__":
    main()



from __future__ import annotations

import logging
import os

lgr = logging.getLogger(__name__)
DEBUG = bool(os.environ.get("DEBUG", False))
lgr.setLevel(logging.DEBUG if DEBUG else logging.INFO)

import bids
from jsonschema._utils import Unset
from jsonschema.exceptions import ValidationError

from . import schema


class BIDSJSONError(ValidationError):
    # class to represent BIDS metadata error
    pass


class BIDSFileError(ValidationError):
    # class to represent error of BIDS file missing or unexpected
    pass


def validate(bids_layout: bids.BIDSLayout, **entities: dict[str, str | list]):
    # validates the data specified by entities using the schema present in the `.forbids` folder

    ref_layout = bids.BIDSLayout(os.path.join(bids_layout.root, schema.FORBIDS_SCHEMA_FOLDER), validate=False)

    # get sidecars for the session or ones factored at a higher level
    ref_sidecars = ref_layout.get(session=[entities.get("session"), None], extension=".json")

    all_sidecars = bids_layout.get(extension=".json", **entities)

    subjects = bids_layout.get_subject(subject=entities.pop("subject"))

    is_multisession = len(bids_layout.get_session())
    is_session_specific = len(ref_layout.get_session())
    if is_multisession:
        lgr.info("The dataset is multi-session.")

    all_sidecars = bids_layout.get(
        extension=".json",
        subject=subjects,
        session=entities["session"],
    )

    for ref_sidecar in ref_sidecars:
        lgr.info("validating %s", str(ref_sidecar.relpath))
        # load the schema
        sidecar_schema = ref_sidecar.get_dict()
        bidsfile_constraints = sidecar_schema.pop("bids", dict())
        query_entities = ref_sidecar.entities.copy()

        for entity in schema.ALT_ENTITIES:
            if entity not in query_entities:
                query_entities[entity] = bids.layout.Query.NONE
        validator = schema.get_validator(sidecar_schema)

        for subject in subjects:
            query_entities["subject"] = subject
            if is_session_specific:
                sessions = [query_entities["session"]]
            else:
                sessions = bids_layout.get_session(subject=subject, session=entities["session"]) or [
                    bids.layout.Query.NONE
                ]

            for session in sessions:
                lgr.info("validating sub-%s %s", subject, "ses-" + session if isinstance(session, str) else "")
                query_entities["session"] = session
                non_null_entities = {k: v for k, v in query_entities.items() if not isinstance(v, bids.layout.Query)}
                expected_sidecar = bids_layout.build_path(non_null_entities, absolute_paths=False)

                lgr.debug(query_entities)

                session_instrument_tags = {
                    k: bids_layout.__getattr__(f"get_{k}")(subject=subject, session=session)[0]
                    for k in bidsfile_constraints["instrument_tags"]
                }
                session_instrument_key = schema.get_instrument_key(
                    session_instrument_tags, bidsfile_constraints["instrument_tags"]
                )

                sidecars_to_validate = bids_layout.get(**query_entities)

                if not sidecars_to_validate:
                    if not bidsfile_constraints.get("optional", False):
                        if session_instrument_key in bidsfile_constraints.get("required_for_instruments", []):
                            yield BIDSFileError(f"{expected_sidecar}", "no match")
                            continue  # no point going further
                    else:
                        lgr.info(f"optional {ref_sidecar.relpath} not present for sub-{subject} ses-{session}")

                num_sidecars = len(sidecars_to_validate)
                min_runs = bidsfile_constraints.get("min_runs", 0)
                max_runs = bidsfile_constraints.get("max_runs", 1e10)

                if num_sidecars < min_runs:
                    yield BIDSFileError(
                        f"Expected at least {min_runs} runs for {expected_sidecar}, found {num_sidecars}"
                    )
                elif num_sidecars > max_runs:
                    yield BIDSFileError(
                        f"Expected at most {max_runs} runs for {expected_sidecar}, found {num_sidecars}"
                    )

                for sidecar in sidecars_to_validate:
                    if sidecar in all_sidecars:
                        all_sidecars.remove(sidecar)
                    else:
                        lgr.error("an error occurred")
                    lgr.debug("validating %s", sidecar.relpath)
                    sidecar_data = schema.prepare_metadata(sidecar, bidsfile_constraints["instrument_tags"])
                    yield from add_path_note_to_error(validator, sidecar_data, sidecar.relpath)
    for extra_sidecar in all_sidecars:
        yield BIDSFileError(f"Unexpected BIDS file {extra_sidecar.relpath}")


def add_path_note_to_error(validator, sidecar_data, filepath):
    # add the path to the json files that triggers the error
    # for better reporting in process_validation

    for error in validator.iter_errors(sidecar_data):
        error.add_note(filepath)
        yield error


def process_validation(layout, subject, session):
    # run validation on the BIDS layout and specified subject/session
    # format errors for not-to-verbose pretty printing

    no_error = True
    for error in validate(layout, subject=subject, session=session):
        no_error = False

        formatted_message = error.message
        if len(error.path) == 0 and not isinstance(error.instance, Unset) and not error.relative_schema_path[0] == 'required':
            formatted_message = f"non-existing schema for instrument {error.instance['__instrument__']}"

        lgr.error(
            "%s %s %s : %s",
            error.__class__.__name__,
            error.__notes__[0] if hasattr(error, "__notes__") else "",
            ".".join([str(e) for e in error.absolute_path]),
            formatted_message,
        )
        lgr.debug(error)
    if no_error:
        lgr.info("The dataset was successfully checked as compliant to the protocol.")
    else:
        lgr.error("The dataset failed to comply to the protocol.")
    return no_error



from __future__ import annotations

from apischema.json_schema import deserialization_schema

from forbids.schema import tagpreset2type


def test_tagpreset2type():
    typ_sc = deserialization_schema(tagpreset2type("any", "*", 10.0))
    typ_sc.pop("$schema")
    assert typ_sc == {"type": "number"}

    typ_sc = deserialization_schema(tagpreset2type("near", "~=.05", 10.0))
    typ_sc.pop("$schema")
    assert typ_sc == {"type": "number", "minimum": 9.95, "maximum": 10.05}

    typ_sc = deserialization_schema(tagpreset2type("eq", "=", 10.0))
    typ_sc.pop("$schema")
    assert typ_sc == {"type": "number", "const": 10}

    typ_sc = deserialization_schema(tagpreset2type("regex", "r^[aA]{3}[0-4]$", "AaA3"))
    typ_sc.pop("$schema")
    assert typ_sc == {
        "type": "string",
        "pattern": "^[aA]{3}[0-4]$",
    }


def test_struct2schemaprops():
    pass
